import asyncio
import os
import logging

from nekoton import *

FORMAT = '%(levelname)s %(name)s %(asctime)-15s %(filename)s:%(lineno)d %(message)s'
logging.basicConfig(format=FORMAT)
logging.getLogger().setLevel(logging.DEBUG)

dirname = os.path.dirname(__file__)

# Address
assert (not Address.validate("totally invalid address"))
assert (Address.validate("0:d84e969feb02481933382c4544e9ff24a2f359847f8896baa86c501c3d1b00cf"))
my_addr = Address('-1:d84e969feb02481933382c4544e9ff24a2f359847f8896baa86c501c3d1b00cf')
assert (my_addr.workchain == -1 and len(my_addr.account) == 32)
my_addr.workchain = 0
assert (my_addr.workchain == 0)

# Cells
cell1 = Cell()
assert (len(Cell().repr_hash) == 32)
assert (Cell() == Cell())

complex_cell_abi = [
    ("first", AbiUint(32)),
    ("second", AbiBool()),
]
complex_cell = Cell.build(
    abi=complex_cell_abi,
    value={
        "first": 123,
        "second": True,
    }
)
assert (Cell.decode(complex_cell.encode('base64'), 'base64') == complex_cell)
assert (Cell.from_bytes(complex_cell.to_bytes()) == complex_cell)

decoded = complex_cell.unpack(abi=complex_cell_abi)
assert (decoded["first"] == 123)
assert (decoded["second"] is True)

# Crypto
some_pubkey = PublicKey('7b671b6bfd43e306d4accb46113a871e66b30cc587a57635766a2f360ee831c6')
print(some_pubkey)
assert some_pubkey == PublicKey.from_int(55816654881951532897500201008042388765613920635159317416907795053873153454534)
assert (some_pubkey.to_int() == 55816654881951532897500201008042388765613920635159317416907795053873153454534)

assert PublicKey('00005a641f7deda1442badd9ed761dd4e948580c7d7f36b3f858ab26b1af6fa1') == PublicKey.from_int(
    623856482362781547816624847020167340480982121355398289209492515375247265)
assert (PublicKey('00005a641f7deda1442badd9ed761dd4e948580c7d7f36b3f858ab26b1af6fa1').to_int()
        == 623856482362781547816624847020167340480982121355398289209492515375247265)

cell_with_pubkey_abi = [
    ("pubkey", AbiUint(256)),
    ("not_pubkey", AbiUint(256)),
]
cell_with_pubkey = Cell.build(
    abi=cell_with_pubkey_abi,
    value={
        "pubkey": some_pubkey,
        "not_pubkey": 123123,
    }
)
decoded_cell_with_pubkey = cell_with_pubkey.unpack(cell_with_pubkey_abi)
assert PublicKey.from_int(decoded_cell_with_pubkey['pubkey']) == some_pubkey
assert decoded_cell_with_pubkey['not_pubkey'] == 123123

seed = Bip39Seed.generate()
assert (seed.word_count == 12)
print(seed)

keypair0 = seed.derive()
assert (len(keypair0.public_key.to_bytes()) == 32)
assert (seed.derive() == seed.derive(path=Bip39Seed.path_for_account(0)))

keypair1 = seed.derive(path=Bip39Seed.path_for_account(1))
assert (keypair0 != keypair1)
assert (len(keypair1.public_key.encode('hex')) == 64)

# Abi
with open(os.path.join(dirname, 'wallet.abi.json'), 'r') as json:
    abi = ContractAbi(json.read())

assert (abi.abi_version == AbiVersion(2, 3))
assert (abi.get_function("non-existing") is None)
assert (abi.get_event("non-existing") is None)

send_transaction_func = abi.get_function("sendTransaction")
assert (send_transaction_func is not None)

send_transaction_input = {
    "dest": my_addr,
    "value": Tokens('1.5'),
    "bounce": False,
    "flags": 3,
    "payload": Cell(),
}
body_cell = send_transaction_func.encode_internal_input(send_transaction_input)
print(body_cell)
assert (body_cell != Cell())

internal_msg = send_transaction_func.encode_internal_message(
    input=send_transaction_input,
    value=Tokens('1.5'),
    bounce=False,
    dst=my_addr,
)
print(internal_msg)
assert (internal_msg.state_init is None)
assert (internal_msg.body == body_cell)
assert (isinstance(internal_msg.header, InternalMessageHeader))

unpacked_body = body_cell.unpack(abi=[
    ("function_id", AbiUint(32)),
    ("dest", AbiAddress()),
    ("value", AbiUint(128)),
    ("bounce", AbiBool()),
    ("flags", AbiUint(8)),
    ("payload", AbiCell()),
])
decoded_body = send_transaction_func.decode_input(message_body=body_cell, internal=True)

for field, item in send_transaction_input.items():
    if isinstance(item, Tokens):
        assert (Tokens.from_nano(unpacked_body[field]) == item)
        assert (Tokens.from_nano(decoded_body[field]) == item)
    else:
        assert (unpacked_body[field] == item)
        assert (decoded_body[field] == item)

unsigned_body = send_transaction_func.encode_external_input(send_transaction_input, public_key=None, address=my_addr)
assert (unsigned_body.sign(keypair0, signature_id=None) == unsigned_body.with_signature(
    keypair0.sign(unsigned_body.hash, signature_id=None)))

unsigned_message = send_transaction_func.encode_external_message(my_addr, send_transaction_input, public_key=None)
print(unsigned_message.without_signature())
external_msg, _ = unsigned_message.without_signature().split()
assert (len(external_msg.hash) == 32)
assert (isinstance(external_msg.header, ExternalInMessageHeader))

depool_abi = ContractAbi.from_file(os.path.join(dirname, 'depool.abi.json'))

tokens = Tokens('10.123456789')
assert (2 * tokens / 2) == tokens


# Subscriptions
async def main():
    clock = Clock()

    transport = JrpcTransport(endpoint="https://jrpc.everwallet.net")
    await transport.check_connection()
    signature_id = await transport.get_signature_id()
    assert (signature_id is None)

    config = await transport.get_blockchain_config()
    assert (config.contains_param(0))
    assert (config.config_address == Address("-1:5555555555555555555555555555555555555555555555555555555555555555"))
    assert (config.elector_address == Address("-1:3333333333333333333333333333333333333333333333333333333333333333"))

    account = await transport.get_account_state(my_addr)
    print(account)
    assert (account is not None)
    assert (account.status == AccountStatus.Active)
    assert (not account.balance.is_zero)
    assert (account.state_init.code is not None)

    executor = TransactionExecutor(config, check_signature=False)
    tx, new_state = executor.execute(unsigned_message.with_fake_signature(), account)
    assert (not tx.aborted)
    assert (tx.compute_phase.exit_code == 0)
    assert (new_state is not None)

    depool_addr = Address("0:d9cf3648c1c9436785ed628d5d83a66853eb85feb94a9cfed6239056a32cc149")
    depool_state = await transport.get_account_state(depool_addr)

    depool_info = depool_abi.get_function("getDePoolInfo").call(depool_state, input={})
    assert (depool_info.exit_code == 0)
    assert (depool_info.output is not None)

    stake_accept_tx = await transport.get_transaction(
        bytes.fromhex("60a311aa3e3c1f30deb3010bb09d0079713ab1b0af07f5fd2ca87f5b282912a4"))
    print(stake_accept_tx)
    on_stake_accept_func = depool_abi.get_function("onStakeAccept")
    parsed_stake_accept = on_stake_accept_func.decode_transaction(stake_accept_tx)
    assert (parsed_stake_accept.input['queryId'] == 93)
    assert (parsed_stake_accept.output == {})

    await transport.trace_transaction(stake_accept_tx).wait()

    full_parsed_stake_accept = depool_abi.decode_transaction(stake_accept_tx)
    assert (full_parsed_stake_accept.function == on_stake_accept_func)
    assert (len(full_parsed_stake_accept.events) == 1)
    assert (full_parsed_stake_accept.events[0][0] == depool_abi.get_event("RoundStakeIsAccepted"))

    code_hash = bytes.fromhex("7d0996943406f7d62a4ff291b1228bf06ebd3e048b58436c5b70fb77ff8b4bf2")
    addresses = await transport.get_accounts_by_code_hash(code_hash, limit=10)
    assert (len(addresses) == 10)

    transactions = await transport.get_transactions(my_addr, limit=10)
    assert (len(transactions) == 10)
    for tx in transactions:
        assert (len(tx.hash) == 32)
        assert tx.has_in_msg
        assert (tx.type == TransactionType.Ordinary)

    latest_tx = transactions[0]
    fetched_tx = await transport.get_transaction(latest_tx.hash)
    assert (latest_tx == fetched_tx)
    fetched_next_tx = await transport.get_dst_transaction(latest_tx.in_msg_hash)
    assert (fetched_next_tx == fetched_tx)

    async with transport.account_states(my_addr) as states:
        # There is always at least one iteration with the current state
        async for state in states:
            if state is not None:
                print(my_addr, state.balance)
            else:
                print(my_addr, 'account not exists')

            break  # for tests

    async with transport.account_transactions(config.elector_address) as batches:
        async for batch, batch_info in batches:
            print(batch)
            assert len(batch) > 0
            break

    deep_tx_hash = bytes.fromhex('a82e6603dec13499dd43486203b3304122ae89c13a80b189eef8976834cba413')
    deep_tx_nodes = 0
    async for transaction in transport.trace_transaction(deep_tx_hash, yield_root=True):
        deep_tx_nodes += 1
        print(transaction)
    assert deep_tx_nodes == 12


if __name__ == "__main__":
    asyncio.run(main())



from .filters import *


class Id(HashFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "id"
        HashFilter.__init__(field)
        OrderBy.__init__(field)


class BlockId(HashFilter):
    @classmethod
    def __init__(cls):
        HashFilter.__init__("block_id")


class BodyHash(HashFilter):
    @classmethod
    def __init__(cls):
        HashFilter.__init__("body_hash")


class IhrDisabled(BoolFilter):
    @classmethod
    def __init__(cls):
        BoolFilter.__init__("ihr_disabled")


class IhrFee(TokensFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "ihr_fee"
        TokensFilter.__init__(field)
        OrderBy.__init__(field)


class ImportFee(TokensFilter):
    @classmethod
    def __init__(cls):
        field = "import_fee"
        TokensFilter.__init__(field)
        OrderBy.__init__(field)


class Bounce(BoolFilter):
    @classmethod
    def __init__(cls):
        BoolFilter.__init__("bounce")


class Bounced(BoolFilter):
    @classmethod
    def __init__(cls):
        BoolFilter.__init__("bounced")


class Src(AddressFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "src"
        AddressFilter.__init__(field)
        OrderBy.__init__(field)


class SrcWorkchainId(IntFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "src_workchain_id"
        IntFilter.__init__(field)
        OrderBy.__init__(field)


class Dst(AddressFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "dst"
        AddressFilter.__init__(field)
        OrderBy.__init__(field)


class DstWorkchainId(IntFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "dst_workchain_id"
        IntFilter.__init__(field)
        OrderBy.__init__(field)


class MsgType(MessageTypeFilter):
    @classmethod
    def __init__(cls):
        MessageTypeFilter.__init__("msg_type")


class CreatedAt(IntFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "created_at"
        IntFilter.__init__(field)
        OrderBy.__init__(field)


class CreatedLt(IntAsStringFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "created_lt"
        IntAsStringFilter.__init__(field)
        OrderBy.__init__(field)



from .filters import *


class WorkchainId(IntFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "workchain_id"
        IntFilter.__init__(field)
        OrderBy.__init__(field)


class Id(HashFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "id"
        HashFilter.__init__(field)
        OrderBy.__init__(field)


class Aborted(BoolFilter):
    @classmethod
    def __init__(cls):
        BoolFilter.__init__("aborted")


class AccountAddr(AddressFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "account_addr"
        AddressFilter.__init__(field)
        OrderBy.__init__(field)


class BalanceDelta(TokensFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "balance_delta"
        TokensFilter.__init__(field)
        OrderBy.__init__(field)


class BlockId(HashFilter):
    @classmethod
    def __init__(cls):
        HashFilter.__init__("block_id")


class CreditFirst(BoolFilter):
    @classmethod
    def __init__(cls):
        BoolFilter.__init__("credit_first")


class Destroyed(BoolFilter):
    @classmethod
    def __init__(cls):
        BoolFilter.__init__("destroyed")


class Now(IntFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "now"
        IntFilter.__init__(field)
        OrderBy.__init__(field)


class Lt(IntAsStringFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "lt"
        IntAsStringFilter.__init__(field)
        OrderBy.__init__(field)


class PrevTransHash(HashFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "prev_trans_hash"
        HashFilter.__init__(field)
        OrderBy.__init__(field)


class PrevTransLt(IntAsStringFilter, OrderBy):
    @classmethod
    def __init__(cls):
        field = "prev_trans_lt"
        IntAsStringFilter.__init__(field)
        OrderBy.__init__(field)


class OldHash(HashFilter):
    @classmethod
    def __init__(cls):
        HashFilter.__init__("old_hash")


class NewHash(HashFilter):
    @classmethod
    def __init__(cls):
        HashFilter.__init__("new_hash")


class TrType(TransactionTypeFilter):
    @classmethod
    def __init__(cls):
        TransactionTypeFilter.__init__("tr_type")



from typing import Iterable as _Iterable

from nekoton import Address as _Address, Tokens as _Tokens, TransactionType as _TransactionType, MessageType as _MessageType, GqlExprPart


class OrderBy:
    @classmethod
    def __init__(cls, path: str):
        cls._field_path = path

    def asc(self) -> GqlExprPart:
        return GqlExprPart('{{path:"{}",direction:ASC}}'.format(self._field_path))

    def desc(self) -> GqlExprPart:
        return GqlExprPart('{{path:"{}",direction:DESC}}'.format(self._field_path))


class BoolFilter:
    @classmethod
    def __init__(cls, field: str):
        cls._field = field

    def _op(self, value: bool) -> GqlExprPart:
        if value:
            return GqlExprPart('{}:{{eq:true}}'.format(self._field))
        else:
            return GqlExprPart('{}:{{eq:false}}'.format(self._field))

    def __invert__(self) -> GqlExprPart:
        return self._op(False)

    def __eq__(self, other: bool) -> GqlExprPart:
        return self._op(other)

    def __ne__(self, other: bool) -> GqlExprPart:
        return self._op(not other)


class IntFilter:
    @classmethod
    def __init__(cls, field: str):
        cls._field = field

    def any_of(self, values: _Iterable[int]) -> GqlExprPart:
        return self._multi_op('in', values)

    def not_any_of(self, values: _Iterable[int]) -> GqlExprPart:
        return self._multi_op('notIn', values)

    def _op(self, op: str, value: int) -> GqlExprPart:
        return GqlExprPart('{}:{{{}:{}}}'.format(self._field, op, value))

    def _multi_op(self, op: str, values: _Iterable[int]) -> GqlExprPart:
        return GqlExprPart('{}:{{{}:[{}]}}'.format(self._field, op, ','.join(map(str, values))))

    def __eq__(self, other: int) -> GqlExprPart:
        return self._op("eq", other)

    def __ne__(self, other: int) -> GqlExprPart:
        return self._op("ne", other)

    def __gt__(self, other: int) -> GqlExprPart:
        return self._op("gt", other)

    def __lt__(self, other: int) -> GqlExprPart:
        return self._op("lt", other)

    def __le__(self, other: int) -> GqlExprPart:
        return self._op("le", other)


class StringFilter:
    @classmethod
    def __init__(cls, field: str):
        cls._field = field

    def any_of(self, values: _Iterable[str]) -> GqlExprPart:
        return self._multi_op('in', values)

    def not_any_of(self, values: _Iterable[str]) -> GqlExprPart:
        return self._multi_op('notIn', values)

    def _op(self, op: str, value: str) -> GqlExprPart:
        return GqlExprPart('{}:{{{}:"{}"}}'.format(self._field, op, value))

    def _multi_op(self, op: str, values: _Iterable[str]) -> GqlExprPart:
        values = '","'.join(values)
        if not values:
            return GqlExprPart('{}:{{{}:[]}}'.format(self._field, op))
        else:
            return GqlExprPart('{}:{{{}:["{}"]}}'.format(self._field, op, values))

    def __eq__(self, other: str) -> GqlExprPart:
        return self._op("eq", other)

    def __ne__(self, other: str) -> GqlExprPart:
        return self._op("ne", other)

    def __gt__(self, other: str) -> GqlExprPart:
        return self._op("gt", other)

    def __lt__(self, other: str) -> GqlExprPart:
        return self._op("lt", other)

    def __le__(self, other: str) -> GqlExprPart:
        return self._op("le", other)


class IntAsStringFilter(StringFilter):
    def any_of(self, values: _Iterable[int | str]) -> GqlExprPart:
        return self._multi_op('in', map(str, values))

    def not_any_of(self, values: _Iterable[int | str]) -> GqlExprPart:
        return self._multi_op('notIn', map(str, values))

    def __eq__(self, other: int | str) -> GqlExprPart:
        return self._op("eq", str(other))

    def __ne__(self, other: int | str) -> GqlExprPart:
        return self._op("ne", str(other))

    def __gt__(self, other: int | str) -> GqlExprPart:
        return self._op("gt", str(other))

    def __lt__(self, other: int | str) -> GqlExprPart:
        return self._op("lt", str(other))

    def __le__(self, other: int | str) -> GqlExprPart:
        return self._op("le", str(other))


class TokensFilter(StringFilter):
    @staticmethod
    def __convert(value: _Tokens | int | str) -> str:
        if isinstance(value, _Tokens):
            return str(value.to_nano())
        else:
            return str(value)

    def any_of(self, values: _Iterable[_Tokens | int | str]) -> GqlExprPart:
        return self._multi_op('in', map(TokensFilter.__convert, values))

    def not_any_of(self, values: _Iterable[_Tokens | int | str]) -> GqlExprPart:
        return self._multi_op('notIn', map(TokensFilter.__convert, values))

    def __eq__(self, other: _Tokens | int | str) -> GqlExprPart:
        return self._op("eq", TokensFilter.__convert(other))

    def __ne__(self, other: _Tokens | int | str) -> GqlExprPart:
        return self._op("ne", TokensFilter.__convert(other))

    def __gt__(self, other: _Tokens | int | str) -> GqlExprPart:
        return self._op("gt", TokensFilter.__convert(other))

    def __lt__(self, other: _Tokens | int | str) -> GqlExprPart:
        return self._op("lt", TokensFilter.__convert(other))

    def __le__(self, other: _Tokens | int | str) -> GqlExprPart:
        return self._op("le", TokensFilter.__convert(other))


class HashFilter(StringFilter):
    @staticmethod
    def __convert(value: bytes | str) -> str:
        if isinstance(value, (bytes, bytearray)):
            return value.hex()
        else:
            return value

    def any_of(self, values: _Iterable[bytes | str]) -> GqlExprPart:
        return self._multi_op('in', map(HashFilter.__convert, values))

    def not_any_of(self, values: _Iterable[bytes | str]) -> GqlExprPart:
        return self._multi_op('notIn', map(HashFilter.__convert, values))

    def __eq__(self, other: bytes | str) -> GqlExprPart:
        return self._op("eq", HashFilter.__convert(other))

    def __ne__(self, other: bytes | str) -> GqlExprPart:
        return self._op("ne", HashFilter.__convert(other))

    def __gt__(self, other: bytes | str) -> GqlExprPart:
        return self._op("gt", HashFilter.__convert(other))

    def __lt__(self, other: bytes | str) -> GqlExprPart:
        return self._op("lt", HashFilter.__convert(other))

    def __le__(self, other: bytes | str) -> GqlExprPart:
        return self._op("le", HashFilter.__convert(other))


class AddressFilter(StringFilter):
    def any_of(self, values: _Iterable[_Address | str]) -> GqlExprPart:
        return self._multi_op('in', map(str, values))

    def not_any_of(self, values: _Iterable[_Address | str]) -> GqlExprPart:
        return self._multi_op('notIn', map(str, values))

    def __eq__(self, other: _Address | str) -> GqlExprPart:
        return self._op("eq", str(other))

    def __ne__(self, other: _Address | str) -> GqlExprPart:
        return self._op("ne", str(other))

    def __gt__(self, other: _Address | str) -> GqlExprPart:
        return self._op("gt", str(other))

    def __lt__(self, other: _Address | str) -> GqlExprPart:
        return self._op("lt", str(other))

    def __le__(self, other: _Address | str) -> GqlExprPart:
        return self._op("le", str(other))


class TransactionTypeFilter(IntFilter):
    def any_of(self, values: _Iterable[_TransactionType | int]) -> GqlExprPart:
        return self._multi_op('in', map(int, values))

    def not_any_of(self, values: _Iterable[_TransactionType | str]) -> GqlExprPart:
        return self._multi_op('notIn', map(int, values))

    def __eq__(self, other: _TransactionType | str) -> GqlExprPart:
        return self._op("eq", int(other))

    def __ne__(self, other: _TransactionType | str) -> GqlExprPart:
        return self._op("ne", int(other))

    def __gt__(self, other: _TransactionType | str) -> GqlExprPart:
        return self._op("gt", int(other))

    def __lt__(self, other: _TransactionType | str) -> GqlExprPart:
        return self._op("lt", int(other))

    def __le__(self, other: _TransactionType | str) -> GqlExprPart:
        return self._op("le", int(other))


class MessageTypeFilter(IntFilter):
    def any_of(self, values: _Iterable[_MessageType | int]) -> GqlExprPart:
        return self._multi_op('in', map(int, values))

    def not_any_of(self, values: _Iterable[_MessageType | str]) -> GqlExprPart:
        return self._multi_op('notIn', map(int, values))

    def __eq__(self, other: _MessageType | str) -> GqlExprPart:
        return self._op("eq", int(other))

    def __ne__(self, other: _MessageType | str) -> GqlExprPart:
        return self._op("ne", int(other))

    def __gt__(self, other: _MessageType | str) -> GqlExprPart:
        return self._op("gt", int(other))

    def __lt__(self, other: _MessageType | str) -> GqlExprPart:
        return self._op("lt", int(other))

    def __le__(self, other: _MessageType | str) -> GqlExprPart:
        return self._op("le", int(other))



from typing import List as _List
from nekoton import GqlExprPart

from . import filters
from . import msg
from . import tx


def and_(expressions: str | GqlExprPart | _List[GqlExprPart]) -> GqlExprPart:
    if isinstance(expressions, GqlExprPart):
        return expressions
    if isinstance(expressions, str):
        return GqlExprPart(expressions)
    else:
        return GqlExprPart(','.join(map(str, expressions)))


def or_(expressions: _List[str | GqlExprPart | _List[GqlExprPart]]) -> GqlExprPart:
    last_part = None
    for expr in reversed(expressions):
        element = and_(expr)
        if last_part is not None:
            last_part = '{},OR:{{{}}}'.format(element, last_part)
        else:
            last_part = element

    if last_part is None:
        raise RuntimeError("Empty OR for gql filter")

    return last_part



from os import PathLike
from typing import Any, ClassVar, Optional, List, Tuple, Dict


#########
## ABI ##
#########
# <editor-fold desc="ABI">


class TransactionExecutor:
    """
    Local transaction executor.

    :param config: blockchain config which will be used during execution.
    :param clock: optional clock to modify used timestamp.
    :param check_signature: whether to check signature.
    """

    check_signature: bool
    """Whether to require valid signatures."""

    @classmethod
    def __init__(cls, config: BlockchainConfig, clock: Optional[Clock] = None,
                 check_signature: Optional[bool] = None) -> None: ...

    def execute(
            self,
            message: Message,
            account: Optional[AccountState] = None
    ) -> Tuple[Transaction, Optional[AccountState]]:
        """
        Executes the specified message on account state.

        :param message: message to execute.
        :param account: account state. (`None` for non-existing).
        """
        ...


class ContractAbi:
    """
    Parsed contract ABI.

    :param abi: a string with JSON ABI description.
    """

    @staticmethod
    def from_file(file: str | bytes | PathLike[str] | PathLike[bytes]) -> ContractAbi:
        """Reads ABI from file."""
        ...

    @classmethod
    def __init__(cls, abi: str) -> None: ...

    @property
    def abi_version(self) -> AbiVersion:
        """TVM ABI version."""
        ...

    def get_function(self, name: str) -> Optional[FunctionAbi]:
        """
        Searches for the function ABI with the specified name.

        :param name: function name.
        """
        ...

    def get_event(self, name: str) -> Optional[EventAbi]:
        """
        Searches for the event ABI with the specified name.

        :param name: event name.
        """
        ...

    def encode_init_data(
            self,
            data: Dict[str, Any],
            public_key: Optional[PublicKey] = None,
            existing_data: Optional[Cell] = None
    ) -> Cell:
        """
        Encodes initial contract data using the specified values and public key.

        :param data: initial data values.
        :param public_key: updates public key if specified.
        :param existing_data: updates the specified initial contract data.
        """
        ...

    def decode_init_data(self, data: Cell) -> Tuple[Optional[PublicKey], Dict[str, Any]]:
        """
        Decodes initial contract data using the contract ABI.

        :param data: initial contract data (from TVC).
        """
        ...

    def decode_transaction(self, transaction: Transaction) -> Optional[FunctionCallFull]:
        """
        Decodes function call and events from the specified transaction.

        :param transaction: transaction to decode.
        """
        ...

    def decode_transaction_events(self, transaction: Transaction) -> List[Tuple[EventAbi, Dict[str, Any]]]:
        """
        Decodes only events from the specified transaction.

        :param transaction: transaction to decode.
        """
        ...


class FunctionAbi:
    """Parsed function ABI."""

    @property
    def abi_version(self) -> AbiVersion:
        """TVM ABI version."""
        ...

    @property
    def name(self) -> str:
        """Function name."""
        ...

    @property
    def input_id(self) -> int:
        """Input id."""
        ...

    @property
    def output_id(self) -> int:
        """Output id."""
        ...

    def call(
            self,
            account_state: AccountState,
            input: Dict,
            responsible: Optional[bool] = None,
            clock: Optional[Clock] = None,
    ) -> ExecutionOutput:
        """
        Runs this function as a getter.

        :param account_state: a state of existing account which will be used for execution.
        :param input: function intput.
        :param responsible: whether to run this getter as responsible.
        :param clock: optional clock to modify execution timestamp.
        """
        ...

    def encode_external_message(
            self,
            dst: Address,
            input: Dict[str, Any],
            public_key: Optional[PublicKey],
            state_init: Optional[StateInit] = None,
            timeout: Optional[int] = None,
            clock: Optional[Clock] = None
    ) -> UnsignedExternalMessage:
        """
        Encodes external message using the function ABI.

        :param dst: destination account address.
        :param input: function arguments.
        :param public_key: public key which will be used for signature.
        :param state_init: optional state init.
        :param timeout: expiration timeout.
        :param clock: optional clock to modify used timestamp.
        """
        ...

    def encode_external_input(
            self,
            input: Dict[str, Any],
            public_key: Optional[PublicKey],
            timeout: Optional[int] = None,
            address: Optional[Address] = None,
            clock: Optional[Clock] = None
    ) -> UnsignedBody:
        """
        Encodes external function input using the function ABI.

        :param input: function arguments.
        :param public_key: public key which will be used for signature.
        :param timeout: expiration timeout.
        :param address: destination account address (for ABI 2.3).
        :param clock: optional clock to modify used timestamp.
        """
        ...

    def encode_internal_message(
            self,
            input: Dict[str, Any],
            value: Tokens,
            bounce: bool,
            dst: Address,
            src: Optional[Address] = None,
            state_init: Optional[StateInit] = None,
    ) -> Message:
        """
        Encodes internal message using the function ABI.

        :param input: function arguments.
        :param value: attached amount of nano EVERs.
        :param bounce: whether to return the amount in case of error.
        :param dst: destination account address.
        :param src: source account address.
        :param state_init: optional state init.
        """
        ...

    def encode_internal_input(self, input: Dict[str, Any]) -> Cell:
        """
        Encodes internal function input using the function ABI.

        :param input: function arguments.
        """
        ...

    def decode_transaction(self, transaction: Transaction) -> FunctionCall:
        """
        Decodes transaction as a function call using the function ABI.

        :param transaction: transaction to decode.
        """
        ...

    def decode_input(
            self,
            message_body: Cell,
            internal: bool,
            allow_partial: Optional[bool] = None
    ) -> Dict[str, Any]:
        """
        Decodes message body as input using the function ABI.

        :param message_body: message body to decode.
        :param internal: whether this body is from an internal message.
        :param allow_partial: whether to decode only the prefix of the body.
        """
        ...

    def decode_output(self, message_body: Cell, allow_partial: Optional[bool] = None) -> Dict[str, Any]:
        """
        Decodes message body as output using the function ABI.

        :param message_body: message body to decode.
        :param allow_partial: whether to decode only the prefix of the body.
        """
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class ExecutionOutput:
    @property
    def exit_code(self) -> int:
        """Exit code from the compute phase."""
        ...

    @property
    def output(self) -> Optional[Dict[str, Any]]:
        """Parsed output in case of successful execution."""
        ...


class FunctionCall:
    """Parsed function call."""

    @property
    def input(self) -> Dict[str, Any]:
        """Parsed function input."""
        ...

    @property
    def output(self) -> Dict[str, Any]:
        """Parsed function output."""
        ...


class FunctionCallFull(FunctionCall):
    """Extended parsed function cell."""

    @property
    def events(self) -> List[Tuple[EventAbi, Dict[str, Any]]]:
        """Parsed events"""
        ...

    @property
    def function(self) -> FunctionAbi:
        """ABI object of the parsed function"""
        ...


class EventAbi:
    """Parsed event ABI."""

    @property
    def abi_version(self) -> AbiVersion:
        """TVM ABI version."""
        ...

    @property
    def name(self) -> str:
        """Event name."""
        ...

    @property
    def id(self) -> int:
        """Event id."""
        ...

    def decode_message(self, message: Message) -> Dict[str, Any]:
        """
        Tries to decode event data from the specified message using event ABI.

        :param message: message to decode.
        """
        ...

    def decode_message_body(self, message_body: Cell) -> Dict[str, Any]:
        """
        Tries to decode event data from the specified message body using event ABI.

        :param message_body: message body to decode.
        """
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class SignedExternalMessage(Message):
    """
    External message with an additional expiration param.
    """

    @property
    def expire_at(self) -> int:
        """Expiration unix timestamp."""
        ...

    def split(self) -> Tuple[Message, int]:
        """Splits into inner message and expiration timestamp."""
        ...


class UnsignedExternalMessage:
    """Unsigned external message with function intput."""

    state_init: Optional[StateInit]
    """Optional state init."""

    @property
    def hash(self) -> bytes:
        """A hash to sign."""
        ...

    @property
    def expire_at(self) -> int:
        """Expiration unix timestamp."""
        ...

    def sign(self, keypair: KeyPair, signature_id: Optional[int]) -> SignedExternalMessage:
        """
        Signs function input with the specified keypair and signature id.

        :param keypair: signer keypair.
        :param signature_id: optional signature id.
        """
        ...

    def with_signature(self, signature: Signature) -> SignedExternalMessage:
        """
        Inserts a signature into the body.

        :param signature: a valid ed25519 signature.
        """
        ...

    def with_fake_signature(self) -> SignedExternalMessage:
        """Inserts a fake signature into the body."""
        ...

    def without_signature(self) -> SignedExternalMessage:
        """Creates an input without a signature."""
        ...


class UnsignedBody:
    """Unsigned function input."""

    @property
    def hash(self) -> bytes:
        """A hash to sign."""
        ...

    @property
    def expire_at(self) -> int:
        """Expiration unix timestamp."""
        ...

    def sign(self, keypair: KeyPair, signature_id: Optional[int]) -> Cell:
        """
        Signs function input with the specified keypair and signature id.

        :param keypair: signer keypair.
        :param signature_id: optional signature id.
        """
        ...

    def with_signature(self, signature: Signature) -> Cell:
        """
        Inserts a signature into the body.

        :param signature: a valid ed25519 signature.
        """
        ...

    def with_fake_signature(self) -> Cell:
        """Inserts a fake signature into the body."""
        ...

    def without_signature(self) -> Cell:
        """Creates an input without a signature."""
        ...


class AbiParam:
    """
    Base ABI type.
    """


class AbiUint(AbiParam):
    """
    A class for `uintN` ABI type.

    :param size: uint size in bits.
    """

    @classmethod
    def __init__(cls, size: int) -> None: ...


class AbiInt(AbiParam):
    """
    A class for an `intN` ABI type.

    :param size: int size in bits.
    """

    @classmethod
    def __init__(cls, size: int) -> None: ...


class AbiVarUint(AbiParam):
    """
    A class for `varuintN` ABI type.

    :param size: varuint size in bytes.
    """

    @classmethod
    def __init__(cls, size: int) -> None: ...


class AbiVarInt(AbiParam):
    """
    A class for `varintN` ABI type.

    :param size: varint size in bytes.
    """

    @classmethod
    def __init__(cls, size: int) -> None: ...


class AbiBool(AbiParam):
    """
    A class for a `bool` ABI type.
    """

    @classmethod
    def __init__(cls) -> None: ...


class AbiTuple(AbiParam):
    """
    A class for `(T1, T2, ..., Tn)` ABI type.

    :param items: ABI types of inner values.
    """

    @classmethod
    def __init__(cls, items: List[Tuple[str, AbiParam]]) -> None: ...


class AbiArray(AbiParam):
    """
    A class for an `T[]` ABI type.

    :param value_type: the ABI type of array items.
    """

    @classmethod
    def __init__(cls, value_type: AbiParam) -> None: ...


class AbiFixedArray(AbiParam):
    """
    A class for a `T[N]` ABI type.

    :param value_type: the ABI type of array items.
    :param len: number of array items.
    """

    @classmethod
    def __init__(cls, value_type: AbiParam, len: int) -> None: ...


class AbiCell(AbiParam):
    """
    A class for a `cell` ABI type.
    """

    @classmethod
    def __init__(cls) -> None: ...


class AbiMap(AbiParam):
    """
    A class for an `map(K, V)` ABI type.

    :param key_type: the ABI type of mapping keys.
    :param value_type: the ABI type of mapping values.
    """

    @classmethod
    def __init__(cls, key_type: AbiParam, value_type: AbiParam) -> None: ...


class AbiAddress(AbiParam):
    """
    A class for a `address` ABI type.
    """

    @classmethod
    def __init__(cls) -> None: ...


class AbiBytes(AbiParam):
    """
    A class for a `bytes` ABI type.
    """

    @classmethod
    def __init__(cls) -> None: ...


class AbiFixedBytes(AbiParam):
    """
    A class for a `fixedbytesN` ABI type.

    :param len: number of bytes.
    """

    @classmethod
    def __init__(cls, len: int) -> None: ...


class AbiString(AbiParam):
    """
    A class for `string` ABI type.
    """

    @classmethod
    def __init__(cls) -> None: ...


class AbiToken(AbiParam):
    """
    A class for `token` ABI type.
    """

    @classmethod
    def __init__(cls) -> None: ...


class AbiOptional(AbiParam):
    """
    A class for an `optional(T)` ABI type.

    :param value_type: the ABI type of the inner value.
    """

    @classmethod
    def __init__(cls, value_type: AbiParam) -> None: ...


class AbiRef(AbiParam):
    """
    A class for a `ref(T)` ABI type.

    :param value_type: the ABI type of the inner value.
    """

    @classmethod
    def __init__(cls, value_type: AbiParam) -> None: ...


class AbiVersion:
    """
    TVM ABI version.

    :param major: major version component.
    :param minor: minor version component.
    """

    major: int
    """Major TVM ABI version component."""

    minor: int
    """Minor TVM ABI version component."""

    @classmethod
    def __init__(cls, major: int, minor: int) -> None: ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


# </editor-fold>

############
## MODELS ##
############
# <editor-fold desc="MODELS">


class BlockchainConfig:
    """
    Partially parsed blockchain config.
    """

    @property
    def capabilities(self) -> int:
        """Required software capabilities as integer mask."""
        ...

    @property
    def global_version(self) -> int:
        """Required software version."""
        ...

    @property
    def config_address(self) -> Address:
        """Address of the config contract."""
        ...

    @property
    def elector_address(self) -> Address:
        """Address of the elector contract."""
        ...

    @property
    def minter_address(self) -> Address:
        """Address of the minter contract."""
        ...

    @property
    def fee_collector_address(self) -> Address:
        """Address of the fee collector contract."""
        ...

    def contains_param(self, index: int) -> bool:
        """
        Returns `True` if the config contains the specified param.

        :param index: param index.
        """
        ...

    def get_raw_param(self, index: int) -> Optional[Cell]:
        """
        Returns the corresponding config value as a cell.

        :param index: param index.
        """
        ...


class AccountState:
    """
    A state of an existing account.
    """

    @property
    def storage_used(self) -> StorageUsed:
        """Storage usage statistics."""
        ...

    @property
    def last_paid(self) -> int:
        """The last time when storage phase was executed."""
        ...

    @property
    def due_payment(self) -> Optional[Tokens]:
        """Optional account debt in nano EVERs."""
        ...

    @property
    def last_trans_lt(self) -> int:
        """The logical time of the last transaction."""
        ...

    @property
    def balance(self) -> Tokens:
        """Account balance in nano EVERs."""
        ...

    @property
    def status(self) -> AccountStatus:
        """Account status."""
        ...

    @property
    def state_init(self) -> Optional[StateInit]:
        """StateInit for the active account."""
        ...

    @property
    def frozen_state_hash(self) -> Optional[bytes]:
        """A hash of the last known state for the frozen account."""
        ...


class StorageUsed:
    """
    Account storage stats.
    """

    @property
    def cells(self) -> int:
        """Number of cells occupied by this account."""
        ...

    @property
    def bits(self) -> int:
        """Number of bits occupied by this account."""
        ...

    @property
    def public_cells(self) -> int:
        """Number of public cells (libraries) provided by this account."""
        ...


class Transaction:
    """Blockchain transaction."""

    @property
    def hash(self) -> bytes:
        """Hash of the root cell."""
        ...

    @property
    def type(self) -> TransactionType:
        """Transaction type."""
        ...

    @property
    def account(self) -> bytes:
        """Account part of the address."""
        ...

    @property
    def lt(self) -> int:
        """Logical time when the transaction was created."""
        ...

    @property
    def now(self) -> int:
        """Unix timestamp when the transaction was created."""
        ...

    @property
    def prev_trans_hash(self) -> bytes:
        """A hash of the previous transaction."""
        ...

    @property
    def prev_trans_lt(self) -> int:
        """A logical time of the previous transaction."""
        ...

    @property
    def orig_status(self) -> AccountStatus:
        """Account status before this transaction."""
        ...

    @property
    def end_status(self) -> AccountStatus:
        """Account status after this transaction."""
        ...

    @property
    def total_fees(self) -> Tokens:
        """A total amount of fees in nano EVERs."""
        ...

    @property
    def has_in_msg(self) -> bool:
        """Whether this transaction has an incoming message. (`True` for ordinary transactions)."""
        ...

    @property
    def has_out_msgs(self) -> bool:
        """Whether this transaction has any outgoing message."""
        ...

    @property
    def out_msgs_len(self) -> int:
        """Number of outgoing messages."""
        ...

    @property
    def in_msg_hash(self) -> Optional[bytes]:
        """Hash of the incoming message."""
        ...

    @property
    def credit_first(self) -> bool:
        """Whether the account balance was updated before the credit storage phase."""
        ...

    @property
    def aborted(self) -> bool:
        """Whether this transaction was not successful."""
        ...

    @property
    def destroyed(self) -> bool:
        """Whether the account was destroyed during this transaction."""
        ...

    @property
    def storage_phase(self) -> Optional[TransactionStoragePhase]:
        """Optional storage phase."""
        ...

    @property
    def credit_phase(self) -> Optional[TransactionCreditPhase]:
        """Optional credit phase."""
        ...

    @property
    def compute_phase(self) -> Optional[TransactionComputePhase]:
        """Optional compute phase."""
        ...

    @property
    def action_phase(self) -> Optional[TransactionActionPhase]:
        """Optional action phase."""
        ...

    @property
    def bounce_phase(self) -> Optional[TransactionBouncePhase]:
        """Optional bounce phase."""
        ...

    def get_in_msg(self) -> Message:
        """Loads an internal message."""
        ...

    def get_out_msgs(self) -> List[Message]:
        """Loads outgoing messages."""
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class TransactionStoragePhase:
    """Transaction storage phase."""

    @property
    def storage_fees_collected(self) -> Tokens:
        """Amount of collected storage fees in nano EVERs."""
        ...

    @property
    def storage_fees_due(self) -> Optional[Tokens]:
        """Payed storage debt."""
        ...

    @property
    def status_change(self) -> AccountStatusChange:
        """Account status change during this phase."""
        ...


class TransactionCreditPhase:
    """Transaction credit phase."""

    @property
    def due_fees_collected(self) -> Optional[Tokens]:
        """Amount of collected storage fees in nano EVERs."""
        ...

    @property
    def credit(self) -> Tokens:
        """Increased balance in nano EVERs."""
        ...


class TransactionComputePhase:
    """Transaction compute phase."""

    @property
    def success(self) -> bool: ...

    @property
    def msg_state_used(self) -> bool: ...

    @property
    def account_activated(self) -> bool: ...

    @property
    def gas_fees(self) -> Tokens: ...

    @property
    def gas_used(self) -> int: ...

    @property
    def gas_limit(self) -> int: ...

    @property
    def gas_credit(self) -> Optional[int]: ...

    @property
    def mode(self) -> int: ...

    @property
    def exit_code(self) -> int: ...

    @property
    def exit_arg(self) -> Optional[int]: ...

    @property
    def vm_steps(self) -> int: ...

    @property
    def vm_init_state_hash(self) -> bytes: ...

    @property
    def vm_final_state_hash(self) -> bytes: ...


class TransactionActionPhase:
    """Transaction action phase."""

    @property
    def success(self) -> bool: ...

    @property
    def valid(self) -> bool: ...

    @property
    def no_funds(self) -> bool: ...

    @property
    def status_change(self) -> AccountStatusChange: ...

    @property
    def total_fwd_fees(self) -> Optional[Tokens]: ...

    @property
    def total_action_fees(self) -> Optional[Tokens]: ...

    @property
    def result_code(self) -> int: ...

    @property
    def result_arg(self) -> Optional[int]: ...

    @property
    def total_actions(self) -> int: ...

    @property
    def special_actions(self) -> int: ...

    @property
    def skipped_actions(self) -> int: ...

    @property
    def messages_created(self) -> int: ...

    @property
    def action_list_hash(self) -> bytes: ...


class TransactionBouncePhase:
    @property
    def msg_fees(self) -> Tokens: ...

    @property
    def fwd_fees(self) -> Tokens: ...


class TransactionType:
    Ordinary: ClassVar[TransactionType] = ...
    """Ordinary transaction."""

    Tick: ClassVar[TransactionType] = ...
    """Special TickTock transaction (at the beginning of the block, without incoming message)."""

    Tock: ClassVar[TransactionType] = ...
    """Special Tock transaction (at the end of the block, without incoming message)."""

    @property
    def is_ordinary(self) -> bool:
        """
        Returns `True` if the transaction type is `Ordinary`
        """
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __int__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class AccountStatus:
    """
    Account status.
    """

    Active: ClassVar[AccountStatus] = ...
    """Active (deployed) account."""

    Frozen: ClassVar[AccountStatus] = ...
    """Frozen account. (Has no state, can be unfrozen or deleted)."""

    NotExists: ClassVar[AccountStatus] = ...
    """Account does not exist."""

    Uninit: ClassVar[AccountStatus] = ...
    """Account without a state."""

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __int__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class AccountStatusChange:
    """
    Account status change during transaction phase.
    """

    Deleted: ClassVar[AccountStatusChange] = ...
    """Account was deleted."""

    Frozen: ClassVar[AccountStatusChange] = ...
    """Account was frozen."""

    Unchanged: ClassVar[AccountStatusChange] = ...
    """Account status has not changed."""

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __int__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class Message:
    """
    Blockchain message.
    """

    @staticmethod
    def from_bytes(bytes: bytes) -> Message:
        """
        Decodes message from raw bytes.

        :param bytes: raw bytes with BOC.
        """
        ...

    @staticmethod
    def decode(value: str, encoding: Optional[str] = None) -> Message:
        """
        Decodes the message from the encoded BOC.

        :param value: a string with encoded BOC.
        :param encoding: encoding type. `base64` (default) or `hex`.
        """
        ...

    @property
    def hash(self) -> bytes:
        """The hash of the root message cell."""
        ...

    @property
    def is_external_in(self) -> bool:
        """Whether this message is `ExternalIn`."""
        ...

    @property
    def is_external_out(self) -> bool:
        """Whether this message is `ExternalOut`."""
        ...

    @property
    def is_internal(self) -> bool:
        """Whether this message is `Internal`."""
        ...

    @property
    def type(self) -> MessageType:
        """Message type."""
        ...

    @property
    def header(self) -> MessageHeader:
        """Message header"""
        ...

    @property
    def created_at(self) -> int:
        """A unix timestamp when this message was created. (always 0 for `ExternalIn`)."""
        ...

    @property
    def created_lt(self) -> int:
        """A logical timestamp when this message was created. (always 0 for `ExternalIn`)."""
        ...

    @property
    def src(self) -> Optional[Address]:
        """Source address. (None for ExternalIn)."""
        ...

    @property
    def dst(self) -> Optional[Address]:
        """Destination address. (None for `ExternalOut`)."""
        ...

    @property
    def value(self) -> Tokens:
        """Attached amount of nano EVERs. (always 0 for non `Internal`)."""
        ...

    @property
    def bounced(self) -> bool:
        """Whether this message was bounced."""
        ...

    @property
    def body(self) -> Optional[Cell]:
        """Optional message body."""
        ...

    @property
    def state_init(self) -> Optional[StateInit]:
        """Optional state init."""
        ...

    def encode(self, encoding: Optional[str] = None) -> str:
        """
        Encodes the message into BOC.

        :param encoding: encoding type. `base64` (default) or `hex`.
        """
        ...

    def to_bytes(self) -> bytes:
        """Encodes message into raw bytes."""
        ...

    def build_cell(self) -> Cell:
        """Encodes message into a new cell."""
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class MessageHeader:
    """Base message header."""

    def type(self) -> MessageType:
        """Message type."""
        ...


class InternalMessageHeader(MessageHeader):
    """Internal message header."""

    @property
    def ihr_disabled(self) -> bool: ...

    @property
    def bounce(self) -> bool: ...

    @property
    def bounced(self) -> bool: ...

    @property
    def src(self) -> Address: ...

    @property
    def dst(self) -> Address: ...

    @property
    def value(self) -> Tokens: ...

    @property
    def ihr_fee(self) -> Tokens: ...

    @property
    def fwd_fee(self) -> Tokens: ...

    @property
    def created_at(self) -> int: ...

    @property
    def created_lt(self) -> int: ...


class ExternalInMessageHeader(MessageHeader):
    """External incoming message header."""

    @property
    def dst(self) -> Address:
        """Message destination."""
        ...

    @property
    def import_fee(self) -> Tokens:
        """Import fee in nano EVERs"""
        ...


class ExternalOutMessageHeader(MessageHeader):
    """External outgoing message header."""

    @property
    def src(self) -> Address:
        """Message source."""
        ...

    @property
    def created_at(self) -> int:
        """A unix timestamp when the message was created."""
        ...

    @property
    def created_lt(self) -> int:
        """A logical time when the message was created."""
        ...


class MessageType:
    """Message type."""

    Internal: ClassVar[MessageType] = ...
    """Internal message. (Messages between accounts)."""

    ExternalIn: ClassVar[MessageType] = ...
    """External incoming message. (External calls)."""

    ExternalOut: ClassVar[MessageType] = ...
    """External outgoing message. (Events)."""

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __int__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class StateInit:
    """
    Contract code and data.

    :param code: optional contract code.
    :param data: optional contract data.
    """

    code: Optional[Cell]
    """Optional contract code."""

    data: Optional[Cell]
    """Optional contract data."""

    @staticmethod
    def from_bytes(bytes: bytes) -> StateInit:
        """
        Decodes state init from raw bytes.

        :param bytes: raw bytes with BOC.
        """
        ...

    @classmethod
    def __init__(cls, code: Optional[Cell], data: Optional[Cell]) -> None: ...

    @property
    def code_hash(self) -> Optional[bytes]:
        """Optional code hash."""
        ...

    def set_code_salt(self, salt: Cell):
        """
        Tries to update the code salt.

        :param salt: a cell with code salt.
        """
        ...

    def get_code_salt(self) -> Optional[Cell]:
        """
        Tries to extract the code salt
        """
        ...

    def compute_address(self, workchain: Optional[int] = None) -> Address:
        """
        Computes an address for this StateInit.

        :param workchain: optional workchain. (0 by default).
        """
        ...

    def encode(self, encoding: Optional[str] = None) -> str:
        """
        Encodes the state init into BOC.

        :param encoding: encoding type. `base64` (default) or `hex`.
        """
        ...

    def to_bytes(self) -> bytes:
        """Encodes state init into raw bytes."""
        ...

    def build_cell(self) -> Cell:
        """Creates a new cell with StateInit."""
        ...


class Address:
    """
    Account address (`StdAddr`).

    :param addr: a string with raw address.
    """

    workchain: int
    """Address workchain."""

    @staticmethod
    def validate(addr: str) -> bool:
        """
        Checks whether the specified string is a valid address.

        :param addr: a string with raw address.
        """
        ...

    @staticmethod
    def from_parts(workchain: int, account: bytes) -> Any:
        """
        Creates an address from parts.

        :param workchain: address workchain.
        :param account: hash of the initial state.
        """
        ...

    @classmethod
    def __init__(cls, addr: str) -> None: ...

    @property
    def account(self) -> bytes:
        """Hash of the initial state."""
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class Cell:
    """
    A container with up to 1023 bits of data and up to 4 children.
    """

    @staticmethod
    def from_bytes(bytes: bytes) -> Cell:
        """
        Decodes cell from raw bytes.

        :param bytes: raw bytes with BOC.
        """
        ...

    @staticmethod
    def build(
            abi: List[Tuple[str, AbiParam]],
            value: Dict[str, Any],
            abi_version: Optional[AbiVersion] = None,
    ) -> Cell:
        """
        Packs values into cell using the provided ABI.

        :param abi: ABI structure.
        :param value: a dictionary with corresponding values.
        :param abi_version: optional ABI version.
        """
        ...

    @staticmethod
    def decode(value: str, encoding: Optional[str] = None) -> Cell:
        """
        Decodes the cell from the encoded BOC.

        :param value: a string with encoded BOC.
        :param encoding: encoding type. `base64` (default) or `hex`.
        """
        ...

    @classmethod
    def __init__(cls) -> None: ...

    @property
    def repr_hash(self) -> bytes:
        """Representation hash of the cell."""
        ...

    @property
    def bits(self) -> int:
        """Data length in bits."""
        ...

    @property
    def refs(self) -> int:
        """Number of child references."""
        ...

    def encode(self, encoding: Optional[str] = None) -> str:
        """
        Encodes the cell into BOC.

        :param encoding: encoding type. `base64` (default) or `hex`.
        """
        ...

    def to_bytes(self) -> bytes:
        """Encodes cell into raw bytes."""
        ...

    def unpack(
            self,
            abi: List[Tuple[str, AbiParam]],
            abi_version: Optional[AbiVersion] = None,
            allow_partial: Optional[bool] = None,
    ) -> Dict[str, Any]:
        """
        Unpack values using the provided ABI.

        :param abi: ABI structure.
        :param abi_version: optional ABI version.
        :param allow_partial: whether to unpack only the prefix of the cell. (`False` by default).
        """
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class Tokens:
    """
    Wrapper around native currency.
    """

    @staticmethod
    def from_nano(nano: int) -> Tokens:
        """Wraps amount in nano."""
        ...

    @classmethod
    def __init__(cls, value: str | int):
        """Constructs tokens from decimal or integer value."""
        ...

    @property
    def is_signed(self) -> bool:
        """Returns `True` if the argument has a negative sign and `False` otherwise."""
        ...

    @property
    def is_zero(self) -> bool:
        """Returns `True` if the argument is a zero and `False` otherwise."""
        ...

    def max(self, other: Tokens) -> Tokens:
        """Compares two values numerically and returns the maximum."""
        ...

    def min(self, other: Tokens) -> Tokens:
        """Compares two values numerically and returns the minimum."""
        ...

    def to_nano(self) -> int:
        """Returns underlying value as nano."""
        ...

    def abs(self) -> Tokens: ...

    def __bool__(self) -> bool: ...

    def __int__(self) -> int: ...

    def __add__(self, other: Tokens) -> Tokens: ...

    def __sub__(self, other: Tokens) -> Tokens: ...

    def __mul__(self, other: int) -> Tokens: ...

    def __rmul__(self, other: int) -> Tokens: ...

    def __truediv__(self, other: int) -> Tokens: ...

    def __pos__(self) -> Tokens: ...

    def __neg__(self) -> Tokens: ...

    def __abs__(self) -> Tokens: ...

    def __eq__(self, other: Tokens) -> Any: ...

    def __ge__(self, other: Tokens) -> Any: ...

    def __gt__(self, other: Tokens) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other: Tokens) -> Any: ...

    def __lt__(self, other: Tokens) -> Any: ...

    def __ne__(self, other: Tokens) -> Any: ...


# </editor-fold>

###############
## TRANSPORT ##
###############
# <editor-fold desc="TRANSPORT">


class Transport:
    """Base transport"""

    @property
    def clock(self) -> Clock:
        """Time context."""
        ...

    async def check_connection(self):
        """Checks the connection."""
        ...

    async def send_external_message(self, message: SignedExternalMessage) -> Optional[Transaction]:
        """
        Sends an external message to the network and waits until the transaction.

        :param message: signed external message.
        """
        ...

    async def get_signature_id(self) -> Optional[int]:
        """Fetches signature id for the selected network."""
        ...

    async def get_blockchain_config(self, force: Optional[bool] = None) -> BlockchainConfig:
        """
        Fetches the latest blockchain config.

        :param force: whether to ignore cache.
        """
        ...

    async def get_account_state(self, address: Address) -> Optional[AccountState]:
        """
        Fetches an account state for the specified address.

        :param address: account address.
        """
        ...

    async def get_accounts_by_code_hash(
            self,
            code_hash: bytes,
            continuation: Optional[Address] = None,
            limit: Optional[int] = None,
    ) -> List[Address]:
        """
        Fetches a list of address of accounts with the specified code hash.

        :param code_hash: code hash.
        :param continuation: optional account address from the previous batch.
        :param limit: max number of items in response.
        """
        ...

    async def get_transaction(self, transaction_hash: bytes) -> Optional[Transaction]:
        """
        Fetches the transaction by hash.

        :param transaction_hash: transaction hash.
        """
        ...

    async def get_dst_transaction(self, message_hash: bytes | Message) -> Optional[Transaction]:
        """
        Searches for a transaction by the hash of incoming message.

        :param message_hash: a hash of the incoming message, or the message itself.
        """
        ...

    async def get_transactions(
            self,
            address: Address,
            lt: Optional[int] = None,
            limit: Optional[int] = None,
    ) -> List[Transaction]:
        """
        Fetches a transactions batch for the specified account.

        :param address: account address.
        :param lt: optoinal logical time of the latest transaction.
        :param limit: max number of items in response.
        """
        ...

    def account_states(self, address: Address) -> AccountStatesAsyncIter:
        """
        Returns an async account states iterator.

        :param address: account address.
        """
        ...

    def account_transactions(self, address: Address) -> AccountTransactionsAsyncIter:
        """
        Returns an async account transactions iterator.

        :param address: account address.
        """
        ...

    def trace_transaction(self, transaction_hash: bytes | Transaction, yield_root: bool = False) -> TraceTransaction:
        """
        Returns an async transactions iterator over the transactions tree.

        :param transaction_hash: hash of the root transaction, or the root transaction itself.
        :param yield_root: whether to emit the root transaction.
        """
        ...


class GqlTransport(Transport):
    """
    GraphQl transport.

    :param endpoints: a list of gql endpoints.
    :param clock: optional clock to modify timestamp.
    :param local: whether the connection is with local node.
    """

    @classmethod
    def __init__(
            cls,
            endpoints: List[str],
            clock: Optional[Clock] = None,
            local: Optional[bool] = None,
    ) -> None: ...

    async def query_transactions(
            self,
            filter: str | GqlExprPart | List[GqlExprPart],
            order_by: Optional[str | GqlExprPart | List[GqlExprPart]] = None,
            limit: Optional[int] = None
    ) -> List[Transaction]:
        """
        Transactions GQL query.

        :param filter: filter parts.
        :param order_by: optional orderBy parts.
        :param limit: optional limit.
        """
        ...

    async def query_messages(
            self,
            filter: str | GqlExprPart | List[GqlExprPart],
            order_by: Optional[str | GqlExprPart | List[GqlExprPart]] = None,
            limit: Optional[int] = None
    ) -> List[Transaction]:
        """
        Messages GQL query.

        :param filter: filter parts.
        :param order_by: optional orderBy parts.
        :param limit: optional limit.
        """
        ...


class GqlExprPart:
    """
    GQL query part.

    :param value: part value.
    """

    @classmethod
    def __init__(cls, value: str): ...

    def __str__(self): ...


class JrpcTransport(Transport):
    """
    JRPC transport.

    :param endpoint: JRPC endpoint.
    :param clock: optional clock to modify timestamp.
    """

    @classmethod
    def __init__(cls, endpoint: str, clock: Optional[Clock] = None) -> None: ...


class AccountStatesAsyncIter:
    """
    Async account states iterator.
    """

    async def close(self):
        """
        Closes async iterator.
        """
        ...

    async def __aenter__(self) -> AccountStatesAsyncIter: ...

    async def __aexit__(self, exc_type, exc_val, exc_tb): ...

    def __aiter__(self) -> AccountStatesAsyncIter: ...

    def __anext__(self) -> Optional[AccountState]: ...


class AccountTransactionsAsyncIter:
    """
    Async account transactions iterator.
    """

    async def close(self):
        """
        Closes async iterator.
        """
        ...

    async def __aenter__(self) -> AccountTransactionsAsyncIter: ...

    async def __aexit__(self, exc_type, exc_val, exc_tb): ...

    def __aiter__(self) -> AccountTransactionsAsyncIter: ...

    def __anext__(self) -> Tuple[List[Transaction], TransactionsBatchInfo]: ...


class TraceTransaction:
    """
    Async transactions tree iterator.
    """

    async def close(self):
        """
        Closes async iterator.
        """
        ...

    async def wait(self):
        """
        Waits for the last transaction.
        """
        ...

    async def __aenter__(self) -> TraceTransaction: ...

    async def __aexit__(self, exc_type, exc_val, exc_tb): ...

    def __aiter__(self) -> TraceTransaction: ...

    def __anext__(self) -> Transaction: ...


class TransactionsBatchInfo:
    """
    Account transactions batch range info
    """

    @property
    def min_lt(self) -> int:
        """The lowest logical time in batch."""
        ...

    @property
    def max_lt(self) -> int:
        """The highest logical time in batch."""
        ...


class Clock:
    """
    Time context.

    :param offset: optional offset in milliseconds.
    """

    offset: int
    """Clock offset in milliseconds."""

    @classmethod
    def __init__(cls, offset: Optional[int] = None) -> None: ...

    @property
    def now_sec(self) -> int:
        """Returns current timestamp in seconds."""
        ...

    @property
    def now_ms(self) -> int:
        """Returns current timestamp in milliseconds."""
        ...


# </editor-fold>

############
## CRYPTO ##
############
# <editor-fold desc="CRYPTO">


class PublicKey:
    """
    Ed25519 public key.

    :param value: a string with encoded public key.
    :param encoding: encoding of the value. `hex` (default) or `base64`.
    """

    @staticmethod
    def from_int(int: int) -> PublicKey:
        """
        Tries to construct a public key from integer.

        :param int: integer (max 2^256-1)
        """
        ...

    @staticmethod
    def from_bytes(bytes: bytes) -> PublicKey:
        """
        Tries to construct a public key from raw bytes.

        :param bytes: 32 bytes of public key.
        """
        ...

    @classmethod
    def __init__(cls, value: str, encoding: Optional[str] = None) -> None: ...

    def check_signature(self, data: bytes, signature: Signature, signature_id: Optional[int] = None) -> bool:
        """
        Returns `True` if the signature is correct.

        :param data: signed message.
        :param signature: signature to check.
        :param signature_id: optional signature id.
        """
        ...

    def encode(self, encoding: Optional[str] = None) -> str:
        """
        Encodes public key into string.

        :param encoding: encoding of the value. `hex` (default) or `base64`.
        """
        ...

    def to_bytes(self) -> bytes:
        """Converts public key into raw bytes."""
        ...

    def to_int(self) -> int:
        """Converts public key into integer."""
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class KeyPair:
    """
    Ed25519 key pair.

    :param secret: 32 bytes of secret.
    """

    @staticmethod
    def generate() -> KeyPair:
        """Generates a new keypair."""
        ...

    @classmethod
    def __init__(cls, secret: bytes) -> None: ...

    @property
    def secret_key(self) -> PublicKey:
        """Corresponding secret key."""
        ...

    @property
    def public_key(self) -> PublicKey:
        """Corresponding public key."""
        ...

    def sign(self, data: bytes, signature_id: Optional[int]) -> Signature:
        """
        Signs a hash of the specified data.

        :param data: data to sign.
        :param signature_id: optional signature id.
        """
        ...

    def sign_raw(self, data: bytes, signature_id: Optional[int]) -> Signature:
        """
        Signs data as is.

        :param data: data to sign.
        :param signature_id: optional signature id.
        """
        ...

    def check_signature(self, data: bytes, signature: Signature, signature_id: Optional[int] = None) -> bool:
        """
        Returns `True` if the signature is correct.

        :param data: signed message.
        :param signature: signature to check.
        :param signature_id: optional signature id.
        """
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class Signature:
    """
    Ed25519 signature.

    :param value: a string with encoded signature.
    :param encoding: encoding of the value. `hex` (default) of `base64`.
    """

    @staticmethod
    def from_bytes(bytes: bytes) -> Signature:
        """
        Tries to construct a signature from raw bytes.

        :param bytes: 64 bytes of signature.
        """
        ...

    @classmethod
    def __init__(cls, value: str, encoding: Optional[str] = None) -> None: ...

    def encode(self, encoding: Optional[str] = None) -> str:
        """
        Encodes signature into string.

        :param encoding: encoding of the value. `hex` (default) of `base64`.
        """
        ...

    def to_bytes(self) -> bytes:
        """Converts signature into raw bytes."""
        ...

    def __eq__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __gt__(self, other) -> Any: ...

    def __hash__(self) -> Any: ...

    def __le__(self, other) -> Any: ...

    def __lt__(self, other) -> Any: ...

    def __ne__(self, other) -> Any: ...


class Seed:
    """Base seed."""

    @property
    def word_count(self) -> int:
        """Number of words in phrase."""
        ...


class LegacySeed(Seed):
    """
    Legacy seed phase.

    :param phrase: a string with 24 words.
    """

    @staticmethod
    def generate() -> LegacySeed:
        """Generates a new legacy seed."""
        ...

    @classmethod
    def __init__(cls, phrase: str) -> None: ...

    def derive(self) -> KeyPair:
        """Derives a key pair."""
        ...


class Bip39Seed(Seed):
    """
    BIP39 seed.

    :param phrase: a string with 12 words.
    """

    @staticmethod
    def generate() -> Bip39Seed:
        """Generates a random BIP39 seed."""
        ...

    @staticmethod
    def path_for_account(id: int) -> str:
        """
        Returns a default derivation path for the specified account number.

        :param id: account number.
        """
        ...

    @classmethod
    def __init__(cls, phrase: str) -> None: ...

    def derive(self, path: Optional[str] = None) -> KeyPair:
        """
        Derives a key pair using some derivation path.

        :param path: custom derivation path.
        """
        ...

# </editor-fold>



import nekoton as nt

giver_abi = nt.ContractAbi("""{
    "ABI version": 1,
    "functions": [{
        "name": "sendGrams",
        "inputs": [
            {"name": "dest", "type": "address"},
            {"name": "amount", "type": "uint64"}
        ],
        "outputs": []
    }],
    "events": []
}""")

send_grams = giver_abi.get_function("sendGrams")
assert send_grams is not None


class Giver:
    @classmethod
    def __init__(cls, transport: nt.Transport, address: nt.Address):
        cls._transport = transport
        cls._address = address

    @property
    def address(self) -> nt.Address:
        return self._address

    async def give(self, target: nt.Address, amount: nt.Tokens):
        # Prepare external message
        message = send_grams.encode_external_message(
            self._address,
            input={
                "dest": target,
                "amount": amount,
            },
            public_key=None
        ).without_signature()

        # Send external message
        tx = await self._transport.send_external_message(message)
        if tx is None:
            raise RuntimeError("Message expired")

        # Wait until all transactions are produced
        await self._transport.trace_transaction(tx).wait()



from typing import Optional

import nekoton as nt

wallet_abi = nt.ContractAbi("""{
    "ABI version": 2,
    "version": "2.3",
    "header": ["pubkey", "time", "expire"],
    "functions": [{
        "name": "sendTransaction",
        "inputs": [
            {"name": "dest", "type": "address"},
            {"name": "value", "type": "uint128"},
            {"name": "bounce", "type": "bool"},
            {"name": "flags", "type": "uint8"},
            {"name": "payload", "type": "cell"}
        ],
        "outputs": []
    }],
    "events": []
}""")

send_transaction = wallet_abi.get_function("sendTransaction")
assert send_transaction is not None

wallet_code = nt.Cell.decode(
    'te6cckEBBgEA/AABFP8A9KQT9LzyyAsBAgEgAgMABNIwAubycdcBAcAA8nqDCNcY7UTQgwfXAdcLP8j4KM8WI88WyfkAA3HXAQHDAJqDB9cBURO68uBk3oBA1wGAINcBgCDXAVQWdfkQ8qj4I7vyeWa++COBBwiggQPoqFIgvLHydAIgghBM7mRsuuMPAcjL/8s/ye1UBAUAmDAC10zQ+kCDBtcBcdcBeNcB10z4AHCAEASqAhSxyMsFUAXPFlAD+gLLaSLQIc8xIddJoIQJuZgzcAHLAFjPFpcwcQHLABLM4skB+wAAPoIQFp4+EbqOEfgAApMg10qXeNcB1AL7AOjRkzLyPOI+zYS/')
wallet_data_abi = [
    ('publicKey', nt.AbiUint(256)),
    ('timestamp', nt.AbiUint(64)),
]


class EverWallet:
    @classmethod
    def compute_address(cls, public_key: nt.PublicKey, workchain: int = 0) -> nt.Address:
        return cls.compute_state_init(public_key).compute_address(workchain)

    @staticmethod
    def compute_state_init(public_key: nt.PublicKey) -> nt.StateInit:
        data = nt.Cell.build(abi=wallet_data_abi, value={
            "publicKey": public_key,
            "timestamp": 0,
        })
        return nt.StateInit(wallet_code, data)

    @classmethod
    def __init__(cls, transport: nt.Transport, keypair: nt.KeyPair, workchain: int = 0):
        state_init = cls.compute_state_init(keypair.public_key)

        cls._initialized = False
        cls._transport = transport
        cls._keypair = keypair
        cls._state_init = state_init
        cls._address = state_init.compute_address(workchain)

    @property
    def address(self) -> nt.Address:
        return self._address

    async def send(self, dst: nt.Address, value: nt.Tokens, payload: nt.Cell, bounce: bool = False) -> nt.Transaction:
        state_init = await self.__get_state_init()

        signature_id = await self._transport.get_signature_id()

        external_message = send_transaction.encode_external_message(
            self._address,
            input={
                "dest": dst,
                "value": value,
                "bounce": bounce,
                "flags": 3,
                "payload": payload
            },
            public_key=self._keypair.public_key,
            state_init=state_init
        ).sign(self._keypair, signature_id)

        tx = await self._transport.send_external_message(external_message)
        if tx is None:
            raise RuntimeError("Message expired")
        return tx

    async def get_account_state(self) -> Optional[nt.AccountState]:
        return await self._transport.get_account_state(self._address)

    async def get_balance(self) -> nt.Tokens:
        state = await self.get_account_state()
        if state is None:
            return nt.Tokens(0)
        else:
            return state.balance

    async def __get_state_init(self) -> Optional[nt.StateInit]:
        if self._initialized:
            return None

        account_state = await self.get_account_state()
        if account_state is not None and account_state.status == nt.AccountStatus.Active:
            self._initialized = True
            return None
        else:
            return self._state_init



import asyncio
import logging
from nekoton import *

FORMAT = '%(levelname)s %(name)s %(asctime)-15s %(filename)s:%(lineno)d %(message)s'
logging.basicConfig(format=FORMAT)
logging.getLogger().setLevel(logging.DEBUG)


async def main():
    transport = GqlTransport(endpoints=["mainnet.evercloud.dev/89a3b8f46a484f2ea3bdd364ddaee3a3"])
    await transport.check_connection()

    strange_addr = Address("-1:04f64c6afbff3dd10d8ba6707790ac9670d540f37a9448b0337baa6a5a92acac")

    transactions = await transport.query_transactions([
        gql.tx.AccountAddr() == strange_addr,
        gql.tx.TrType() == TransactionType.Tock
    ], order_by=[
        gql.tx.Lt().desc(),
    ])
    print(transactions)

    messages = await transport.query_messages(gql.or_([
        gql.msg.Src() == strange_addr,
        gql.msg.Dst() == strange_addr,
    ]), order_by=[
        gql.msg.CreatedLt().desc(),
    ])
    print(messages)


asyncio.run(main())
